<?php if (!defined('PmWiki')) exit();

/*
  suffr - administer a server democratically, via Wiki

  Copyright 2007,2011 Kirk Zurell <kirk@zurell.name>

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
  
  *** THIS SOFTWARE IS DESIGNED TO PERMIT EXECUTION OF ARBITRARY CODE
  *** ON COMPUTER SYSTEMS ON WHICH IT IS INSTALLED. WHEN INSTALLED
  *** PROPERLY IT ENJOYS ALL PRIVILEGES OF THE ADMINISTRATOR. IT IS
  *** INTENDED TO MANAGE FILES THAT DETERMINE THE SAFE AND CONSISTENT
  *** OPERATION OF THE SERVER. IT DOES NOT PREVENT MISUSE OF THESE
  *** FILES OR OF THE SYSTEM ITSELF.

*/

/** \file suffr.php
 * \brief manage a wiki and server democratically
 *
 * See http://suffr.info
 *
 * This is a module for PMWiki, a PHP wiki found at www.pmwiki.org. It
 * allows for management of the configuration files of the
 * wiki/website's host system, according to a democratic regime.
 *
 * For more documentation, see:
 *  - the page PmWiki.Suffr
 *  - the source code
 *  - the README and INSTALL
 *
 * See the important information in the suffr.php source file that
 * describes intended effects and hazards of using this software.
 */

$RecipeInfo['suffr']['Version'] = '2011-02-05';


/* CONFIGURATION
   Change these values in your local/config.php if desired.

   These values need to match the $WikiDir created with the PageStore
   class. suffr looks through this directory for Profiles pages.

   For instance, to change $WikiDir and these to use wiki.d/Profiles:
   in config.php:
     $WikiDir = new PageStore('wiki.d/$Group/$FullName');
     //...
     $suffrProfilesDir = "wiki.d/Profiles";
*/

//suffr looks here for Profiles pages
SDV($suffrProfilesDir,"wiki.d/Profiles");

SDV($suffrPagePattern,"/(($GroupPattern)([\\/.])?($NamePattern))/");

SDV($suffrProfilesGroup, "Profiles");
SDV($suffrProfilesGroupHome, "Profiles");

//suffr approves pages when this percentage of users sign off on a
//page
SDV($suffrSuccessThreshold, 50/100);

//misc
$clearPTVs = "";



/************************************
 SUFFR module programming 
 ************************************/


/* Allow users to sign a page */

Markup("suffrpagesignature", 
       'directives', 
       '/\\(:pagesignature:\\)/e',
       "Keep(suffrGenerateSignature());");

//should perform author check.
$suffrPageSignFmt = "<a href='{\$PageUrl}?action=sign'>\$LinkText</a>";
function suffrGenerateSignature()
{
  global $Author,$AuthorGroup,$pagename,$suffrPageSignFmt;
  return MakeLink($pagename,//should be user's profile page?
		  $pagename,
		  'Sign this page',
		  '',
		  $suffrPageSignFmt);
}



/* Sign a page, and check whether page actions should take place */

$HandleActions['sign'] = 'suffrSignPage';

function suffrSignPage($pagename, $auth='read')
{
  global $Author,$AuthorPage;
  global $GroupPattern,$NamePattern;
  global $PageStartFmt, $PageEndFmt;
  global $suffrProfilesDir;
  global $suffrSuccessThreshold;
  global $suffrPagePattern;
  global $suffrProfilesGroup;
  global $suffrProfilesGroupHome;
  global $LinkUrl, $LinkText;
  global $PageUrl;

  global $ScriptUrl, $Group, $Name;


  $signatures = 0; $total = 0;
  $SignResultFmt = "";


  /* calculate md5sum of signed page */
  $orig = ReadPage($pagename);//release?
  $md5sum = md5($orig['text']);
  $action = "";

  /* if we know the author, sign this page on the AuthorPage */
  if(!isset($Author))
    {
      //we don't know the author.
      Abort("Problem: must know author");
    }

  $signcount = 0;
  $page = ReadPage($AuthorPage);
  $new = $page;

  //replace existing invalid signature or add new signature
  // was "/\n?\(:signature (($GroupPattern([\\/.]))?($NamePattern)) [0-9a-f]+ :\)|\$/"
  $new['text'] = preg_replace("/\n?\(:signature $pagename [0-9a-f]+ :\)|\$/",
				  "\n(:signature $pagename $md5sum :)",
				  $new['text'],
				  1,
				  $signcount);
  $SignResultFmt .= "<p>Signatures added/replaced: $signcount</p>";

  //post updated page text.
  $_POST['post'] = 'y';
  PostPage($AuthorPage, $page, $new);


  /* check all Profiles pages (except RecentChanges and home page) */

  //Calculate a handy passphrase based on collected signatures.
  //If user signed, their name gets added.
  //This is a stub implementation for mockup purposes.
  $passphrasetext = array();

  $pdir = @opendir($suffrProfilesDir);
  if (!$pdir)
    {
      //no profiles, abort.
      Abort("No profiles");
    } 

  while ( ($each = @readdir($pdir)) !== false ) 
    {
      //deletes
      syslog(LOG_INFO,$each);
      if(@preg_match('/,del-/', $each) > 0) continue;

      $matches = Array(); 
      preg_match($suffrPagePattern,$each,$matches);
      
      if(count($matches) != 5) continue;

      //eliminate all pages not actual profile pages
      if(@$matches[2] !== $suffrProfilesGroup) continue;
      if(@$matches[4] === "RecentChanges") continue;
      if(@$matches[4] === $suffrProfilesGroupHome) continue;
      if(@$matches[4] === "HomePage") continue;
      //debug, could add "abstain" below.
      if(@$matches[4] === "KirkZurell") continue;

      //count total profiles
      $total++;
      
      //read each page for signatures
      $page = ReadPage($each);

      //abstain?
      ;

      //know names of all members
      $passphrasetext[$page['author']] = ''; //keys for all members.

      //look for signatures
      $signed = @preg_match("/\\(:signature $pagename $md5sum :\\)/",
		$page['text']);			 
      if($signed > 0) {
	//this user signed this page
	$signatures++;
	//they participate in the passphrase
	$passphrasetext[$page['author']] = $page['author'];
      }
      
    }

  @closedir($pdir);


  //debug
  $SignResultFmt .= "<p>sigs $signatures / total $total</p>";



  /* If page is signed by a majority, perform action */


  //need "sex": s dotall x extended (ignore whitespace) e execute
  //note order
  // 1 command 2 params 3 optbody 4 body
  $SuffrPatterns = "/\\(:(log|command1|command|rcfile|script|clearvars|topic|removetopic|cmdsig) \\s+ (.+?):\\)((.+?)\\(:\\1\\s*:\\))?/sex";

  /* Notes from previous version:
   old (:log pattern had a hack to remove returns
   old (:rcfile pattern matched newlines, said to fix?
   old (:script pattern said same.
  */

  /* If enough signatures, implement directives. */
  global $instcount;
  if($total > 0)
    if( ( ($signatures/$total)  > $suffrSuccessThreshold )/* || true*/ )
      {
	//page is authorized to execute.
	$instcount = 0;


	//calculate passphrase
	ksort($passphrasetext);

	//debug
	//$SignResultFmt .= "<p>passphrase array: " . print_r($passphrasetext, true) . "</p>";

	$passphrasetext = implode($passphrasetext); //now string

	//debug
	//$SignResultFmt .= "<p>passphrase source: $passphrasetext</p>";

	$passphrasetext = md5($passphrasetext);

	//debug
	$SignResultFmt .= "<p>passphrase was $passphrasetext</p>";

	//NOTE: todo: must ensure cronological execution.
	
	//implement each directive
	$orig['text'] = preg_replace(
	  $SuffrPatterns,
	  "_suffrLogTransparently(suffrInvokeDirective(PSS('$0'),PSS('$1'),PSS('$2'),PSS('$4'),PSS('$passphrasetext')))",
	  $orig['text']);	


	$SignResultFmt .= "\nActions carried out ($instcount).\n\nSee <a href=\"$ScriptUrl?n=$pagename\">$pagename</a>\n";


	syslog(LOG_INFO,"clearing PTVs");
	
	//clear PTVs after use.
	global $clearPTVs;
	if($clearPTVs != "")
	  {
	    $orig['text'] = preg_replace(
	      "/^($clearPTVs):.+$/", //PTVs
	      "$2:", //empty PTVs
	      $orig['text']);
	    $orig['text'] = preg_replace(
	      "/\(:($clearPTVs):[^:]+:\)/", //PTVs
	      "(:$1::)", //empty PTVs
	      $orig['text']);
	    //clear reset list.
	    $clearPTVs = "";
	  }
	
	syslog(LOG_INFO, "Writing signed page");
	//write results back to page
       	WritePage($pagename,$orig);
	syslog(LOG_INFO, "Wrote signed page");
	//notify members?


	/* Stub: write page signature to Stubs/PageSignatures. This
	   will probably be replaced by some external service. Right
	   now, all it does is show that the page text, sans argument
	   information, is as it was at last signoff (even if
	   arguments changed, directives are still the same).
	*/

	$psigncount = 0;
	$opinion = suffrGetPageOpinion($pagename);
	$pagesigs = ReadPage('Stubs/PageSignatures');
	$pagesigs['text'] = preg_replace("/\n?\(:pagesig1 $pagename [0-9a-f]+ :\)|\$/",
				  "\n(:pagesig1 $pagename {$opinion['cur']} :)",
				  $pagesigs['text'],
				  1,
				  $psigncount);
	WritePage('Stubs/PageSignatures', $pagesigs);
	$SignResultFmt .= "<p>Page signatures added/replaced: $psigncount</p>";


      }

  /* Finish; show results of signature */

  syslog(LOG_INFO,"writing result page");

  SDV($suffrPageSignFmt, "<p><a href=\"$ScriptUrl?n=$pagename\">$pagename</a> is signed</p>$SignResultFmt");
  SDV($suffrHandleSignFmt,Array(&$PageStartFmt, &$suffrPageSignFmt, &$PageEndFmt));
  PrintFmt($pagename,$suffrHandleSignFmt);
}

function _suffrLogTransparently($incoming)
{
  syslog(LOG_INFO,$incoming);
  return $incoming;
}

// log actions into pages

$instcount = 0;
function suffrPMLog($message = 'Error')
{
  global $instcount;
  $logtext = "\n(:log " . @date(DATE_RFC822) . " seq:" . $instcount . " :)\n";
  $logtext .= $message . "\n";
  
  if($EnableDiag and @error_get_last() !== NULL)
     $logtext .=  print_r(@error_get_last(),true);

  $logtext .= "(:log:)";
  return $logtext;
}


/* Invoke each directive during "execution" */
function suffrInvokeDirective($whole, $command="", $parameters="", $body="", $passphrase="")
{
  global $instcount;
  //count each action
  $instcount++;

  syslog(LOG_INFO, "invoked '$command' - '$parameters' ");

  $command = "suffrInvoke" . trim($command);

  if(is_callable($command))
    {
      return call_user_func($command, trim($whole), trim($parameters), $body, trim($passphrase));
    }
  else
    {
      return /*PSS*/ $whole . suffrPMLog("Not able to invoke \"$command\"");
    }
}


//no execution for log yet
function suffrInvokelog($whole, $parameters, $body, $passphrase)
{
  syslog(LOG_INFO,"got to invoke log");
  return '';
}

//commands are visible to ps
function suffrInvokecommand($whole, $parameters, $body, $passphrase)
{
  $outlines = Array();
  $retval = 0;

  global $pagename;
  $command = suffrExpandPTV($parameters);

  //escapeshellarg escapeshellcmd
  //$command = 

  syslog(LOG_INFO, "invoke command $command");

  $last = exec($command,$outlines,$retval);
  //$outlines = shell_exec($command);

  return /*PSS*/ ($whole) . suffrPMLog("Command " 
    . ($retval ? "failed" : "succeeded") 
    . " ($retval):\n" . @implode("\n",$outlines));

}

//invokes command, puts passphrasetext to stdin
function suffrInvokecommand1($whole, $parameters,$body,$passphrase = "")
{
  $outlines = Array();
  $retval = 0;

  $command = suffrExpandPTV($parameters);


  $descriptorspec = array(
    0 => array('pipe','r'), //process reads stdin
    1 => array('pipe','w'), //process writes stdout
    2 => array('pipe','w'), //process writes stderr
			  );

  $myproc = proc_open($command,
	    $descriptorspec,
	    $pipes);

  if(!is_resource($myproc))
    {
      return /*PSS*/ ($whole) . suffrPMLog("Command failed: unable to start process\n");
    }
  

  //write out passphrase
  fwrite($pipes[0], $passphrase);
  fclose($pipes[0]);

  //read in stdout and stderr
  do
    {
      if(($stdoutlive = !feof($pipes[1])) == true)
	$stdoutlines .= fread($pipes[1],1024);
      if(($stderrlive = !feof($pipes[2])) == true)
	$stderrlines .= fread($pipes[2],1024);
    }
  while($stdoutlive || $stderrlive);

  $outlines[] = $stdoutlines . "\nError stream:\n" . $stderrlines;

  fclose($pipes[1]);
  fclose($pipes[2]);

  //$outlines[] = "before\n" . @implode("\n",scandir("/dev/mapper")) . "\n";
  $retval = proc_close($myproc);
  sleep(1); //give other processes a chance.
  //$outlines[] = "after\n" . @implode("\n",scandir("/dev/mapper")) . "\n";

  return /*PSS*/ ($whole) . suffrPMLog("Command1 " 
    . ($retval ? "failed" : "succeeded")
    . " ($retval):\n" . @implode("\n",$outlines));

}


/* Create an "rc"-type file */
function suffrInvokercfile($whole, $parameters, $body='', $passphrase)
{
  $filename = suffrExpandPTV($parameters);
  if($filename == '') return;

  
  //check if file is writable
  if(!is_writable(@dirname($filename)))
    {
     return suffrPMLog('unable to write in directory');
    }


  //create enclosing directory
  $exists = is_dir(@dirname($filename));
  if(!$exists)
    if(!@mkdir(dirname($filename),0777,true)) //recursive
      {
	return /*PSS*/ ($whole) . suffrPMLog('unable to make directory');
      }

  
  //open file
  if(!($handle = @fopen($filename,'w')))
    {
      return /*PSS*/ ($whole) . suffrPMLog('unable to open');
    };


  //flock($handle,LOCK_EX);flock($handle,LOCK_UN);

  //expand variables within body
  $body = suffrExpandPTV($body);


  //write contents
  if(!fwrite($handle,$body))
    $log = "File not written";
  else
    $log = "Copied file";


  fclose($handle);


  return /*PSS*/ ($whole) . suffrPMLog($log);

}


/* Emit body as a script, then invoke it */
function suffrInvokescript($whole, $parameters, $body, $passphrase)
{
  global $pagename;
  $script = suffrExpandPTV($body);

  //find a place to put it.
  $filename = tempnam(sys_get_temp_dir(),'wiki');

  if(!($handle = @fopen($filename,'w')))
    {
      return /*PSS*/ ($whole) . suffrPMLog('unable to create script');
    }

  if(!(fwrite($handle,"#!/bin/bash\n") && fwrite($handle,$script)))
    {
      return /* PSS */ ($whole) . suffrPMLog('unable to write script');
    }

  fclose($handle);

  //umask on create? 
  if(!chmod($filename,0700))
    return /*PSS*/ ($whole) . suffrPMLog("unable to chmod");

  $outlines = Array();
  $last = exec($filename,$outlines,$retval);

  return /*PSS*/ ($whole) . suffrPMLog("Script "
    . ($retval ? "failed" : "succeeded")
    . " ($retval):\n" . implode("\n",$outlines));
  
  //delete?
}


/* Clear Page Text Variables after invoking actions */
function suffrInvokeclearvars($whole, $parameters, $body, $passphrase)
{
  global $clearPTVs;

  $clearPTVs = @preg_replace("/\,/","|", $parameters);

  return ($whole) /*. suffrPMLog("Clearing $clearPTVs")*/ ;
}


/* Create a (different) Wiki topic as an action of a topic */
function suffrInvoketopic($whole, $parameters, $body, $passphrase)
{
  $newtopicname = trim(suffrExpandPTV($parameters));

  syslog(LOG_INFO,"loaded ".print_r($newtopicname,true));

  $newtopic = ReadPage($newtopicname);

  if((@$newtopic['text']) == "")
    $newtopic['text'] = suffrExpandPTV($body);

  WritePage($newtopicname,$newtopic);

  syslog(LOG_INFO,"saved ".print_r($newtopic,true));

  return ($whole) . suffrPMLog("Created topic $newtopicname");
}

/* Remove a Wiki topic as an action of another topic */
function suffrInvokeremovetopic($whole, $parameters, $body, $passphrase)
{
  $topicname = trim(suffrExpandPTV($parameters));

  syslog(LOG_INFO, "removed topic $topicname");

  DeletePage($topicname);

  return ($whole) . suffrPMLog("Removed topic $topicname");
}

/* Delete a topic */
function DeletePage($pagename) {
  global $WikiLibDirs,$WikiDir,$LastModFile;
  $WikiDir->iswrite = 1;
  for($i=0; $i<count($WikiLibDirs); $i++) {
    $wd = &$WikiLibDirs[$i];
    if ($wd->iswrite && $wd->exists($pagename)) break;
  }
  if ($i >= count($WikiLibDirs)) $wd = &$WikiDir;
  $wd->delete($pagename);
  if ($LastModFile && !@touch($LastModFile)) 
    { unlink($LastModFile); touch($LastModFile); fixperms($LastModFile); }
}



/* Replace Page Text Variables in text to be sent to system */
function suffrExpandPTV($text)
{
  global $pagename;

  $result = preg_replace("/\\{\\$:(\\w+?)\\}/e", "PageTextVar(\$pagename, PSS('$1'))", $text);
  $result = preg_replace("/\\{(\\(\\w+\\b.*?\\))\\}/e",
	    "MarkupExpression(\$pagename, PSS('$1'))",
			 $result); 
  return $result;
}


/* "Opinion" system: To assist users in managing their system,
 commands and pages (stripped of argument information) are tested
 against a stored checksum */

/* Get opinion about an executable */
SDV($suffrCmdSigcmd, "/\(:cmdsig \\s+ ([[:xdigit:]]+) \\s+ ([[:alnum:]_\-\/]+) \\s* :\)/ix");
function suffrGetCmdOpinion($pagename,$command) {
  global $suffrCmdSigcmd;
    
  /* Get an opinion on the command executable */
  $opinion = 'Not signed';
  //first item in line, probably should improve this.
  if( preg_match('/[[:alnum:]_\-\/]+/', $command, $matches ) > 0 )
    {
      $cmd = $matches[0];
      // find full path?
      
      // calc MD5sum of executable
      if ( $cmdmd = md5_file($cmd) ) {
	//load opinion page
	$opinions = ReadPage('Stubs/CommandSignatures'); //STUB; external service? AJAX? ESP?
      	preg_match_all($suffrCmdSigcmd, $opinions['text'], $pmatches, PREG_PATTERN_ORDER);
	//return 'hi' . print_r($pmatches, true);
	$cmds = $pmatches[2]; $sigs = $pmatches[1];
	$key = array_search($cmdmd, $sigs);
	if($key !== FALSE) {
	  //cmd matches key on record
	  $opinion = "Signed";
	}
      }
    }
  return $opinion;
}

/* Get opinion about a page (stripped) */
function suffrGetPageOpinion($pagename) {

  global $suffrPageFmt, $suffrMD5Fmt;
  
  $opinion = array('opinion' => "Not reviewed", 'op' => 'n');

  $cur = ReadPage($pagename);
  //strip out value of any pagetextvars
  // (:Country: Transylvania :)
  $varpats = array("/\\(: \\s* ([[:alnum:]]+) \\s* : \\s* (.+) \\s* :\\)/x");
  $varreps = array("(:\\1::)");
  $cur = preg_replace($varpats, $varreps, $cur['text']); 
  
  //sig of static portion
  $cur = md5($cur);
  $opinion['cur'] = $cur;

  //test text sans parameters (PTVs) against stored checksum.
  $sigs = ReadPage('Stubs/PageSignatures');
  preg_match_all( "/\(:pagesig1 \\s+ ($suffrPageFmt) \\s+ ($suffrMD5Fmt) \\s* :\)/x",
    $sigs['text'],
    $matches,
    PREG_PATTERN_ORDER);
  $sigs = $matches[2];
  $key = array_search($cur, $sigs);
  if($key !== FALSE) {
    $opinion['opinion'] = "reviewed"; $opinion['op'] = 'y';
  }

  return $opinion;
}


//Markup('{(', '>{$var}',
//  '/\\{(\\(\\w+\\b.*?\\))\\}/e',
//  "MarkupExpression(\$pagename, PSS('$1'))");

/********************************************
   Markup/Display of Suffr directives. 
 ********************************************/

/* (:signature page md5 :) */
/* Display signatures on profile page, and whether valid or not */
SDV($suffrPageSignValidFmt, "<a href=\"\$LinkUrl\">\$LinkText</a>");
SDV($suffrMD5Fmt, "[[:xdigit:]]+");

Markup("suffrsignature",
       'directives',
       "/\\(:signature (($GroupPattern([\\/.]))?($NamePattern)) ($suffrMD5Fmt) :\\)/e",
       "'<br/>' . Keep(suffrCheckSignature(PSS('$1'),PSS('$5')))");

function suffrCheckSignature($target, $md5sum)
{
  global $pagename;
  
  $page = ReadPage($target);
  $totest = md5($page['text']);

  $valid = ($totest === $md5sum);

  return MakeLink($pagename,
		  PSS($target),
		  NULL,//use target name
		  ($valid ? " (valid)" : " (invalid)"),
		  $suffrPageSignValidFmt,
		  'L');
}




/* (:command exe arg arg arg :) */
Markup('suffrcommand','directives',
  "/\\(:command \\s+ (.+?) \\s* :\\)/ex",
  "Keep(suffrDisplayCommand(\$pagename,PSS('$1')))");
$HTMLStylesFmt['command'] = 'p.command { background-color: #FFFFDD; }';
$HTMLStylesFmt['opinion'] = 'div.opinion { display: block; float:right; padding: 0pt 3pt; background-color: #888866; color: #ffffff; font-size: 5pt; }';

function suffrDisplayCommand($pagename,$command)
{
  $opinion = suffrGetCmdOpinion($pagename,$command);
  return "\n<div class='command'><div class='opinion'>$opinion</div><p class='command'><em>Command:</em> $command</p></div>";
}

/* (:command1 exe arg arg arg :) + passphrase data */
Markup('suffrcommand1','directives',
  "/\\(:command1 \\s+ (.+?) \\s* :\\)/ex",
  "Keep(suffrDisplayCommand1(\$pagename,PSS('$1')))");
$HTMLStylesFmt['command1'] = 'p.command1 { background-color: #FFFFCC; }';

function suffrDisplayCommand1($pagename,$command)
{
  $opinion = suffrGetCmdOpinion($pagename, $command);
  return "\n<div class='command'><p class='command1'><em>Command1:</em> $command</p><div class='opinion'>$opinion</div></div>";
}

/* (:rcfile target/path :) create/update an rc-type file */
Markup('suffrrcfile', 'fulltext', //">{$var}"
  "/\\(:rcfile \\s+ (.+?) \\s* :\\)(.*?)\\(:rcfile \\s* :\\)/seix",
  "Keep(suffrDisplayRCFile(\$pagename, PSS('$1'), PSS('$2')))");
$HTMLStylesFmt['rcfile-legend'] = 'p.rcfile-legend { background-color: #EEFFFF; }';
$HTMLStylesFmt['rcfile'] = 'pre.rcfile { background-color: #DDFFFF; }';

function suffrDisplayRCFile($pagename, $filename, $body)
{
  $output = "\n<p class='rcfile-legend'><em>File:</em> $filename</p>\n";
  $output .= "<pre class='rcfile'>";
  $output .= $body;
  $output .= "</pre>\n";
  return $output;
}

/* (:script:) emit and run script */
Markup('suffrscript', 'fulltext',
       "/\\(:script \\s* :\\)(.*?)\\(:script \\s* :\\)/seix",
  "Keep(suffrDisplayScript(\$pagename, PSS('$1')))");
$HTMLStylesFmt['script-legend'] = 'p.script-legend { background-color: #EEEEFF; }';
$HTMLStylesFmt['script'] = 'pre.script { background-color: #DDDDFF; }';

function suffrDisplayScript($pagename, $body)
{
  $output = "\n<p class='script-legend'><em>Script:</em> #!/bin/bash</p>\n";
  $output .= "<pre class='script'>";
  $output .= $body;
  $output .= "</pre>\n";
  return $output;
}

/* (:log displays logged info recorded from execution of previous
   command */
Markup('suffrlog','fulltext',
  "/\\(:log \\s+ (.*?) \\s* :\\)(.*?)\\(:log \\s* :\\)/seix",
  "Keep(suffrDisplayLog(\$pagename, PSS('$1'), PSS('$2')))");
$HTMLStylesFmt['log-legend'] = 'p.log-legend { background-color: #FFBBFF; }';
$HTMLStylesFmt['log'] = 'pre.log { background-color: #FFDDFF; }';

function suffrDisplayLog($pagename, $logdate, $logtext)
{
  $output = "\n<p class='log-legend'><em>Log:</em> $logdate</p>";
  $output .= "\n<pre class='log'>";
  $output .= $logtext;
  $output .= "</pre>\n";

  // global $MarkupTable;
  // $output = print_r($MarkupTable,true);
  return $output;
}

/* (:tail path/to/file :) displays tail -n50 of a file. */
Markup('suffrtail','directives',
  "/\\(:tail \\s+ (.*?) \\s* :\\)/seix",
  "Keep(suffrDisplayTail(\$pagename, PSS('$1')))");
$HTMLStylesFmt['tail-legend'] = 'p.tail-legend { background-color: #BBFFBB; }';
$HTMLStylesFmt['tail'] = 'pre.tail { background-color: #DDFFDD; }';

function suffrDisplayTail($pagename, $filename)
{
  if(!file_exists($filename))
    $tailtext = "File does not exist...";
  else
    {
      $lastline = exec("tail -n50 $filename",$tailtext,$result);
      $tailtext = implode("\n",$tailtext);
      if($result != 0)
	$tailtext .= "\nfailed.\n";
    }

  $output = "\n<p class='tail-legend'><em>Tail:</em> $filename</p>";
  $output .= "\n<pre class='tail'>";
  $output .= $tailtext;
  $output .= "</pre>\n";
  return $output;
}


/* (:clearvars var var var :) Clears Page Text Vars after all other
   directives are executed. */
Markup('suffrclearptv','directives',
  "/\\(:clearvars \\s+ (.+?) \\s* :\\)/siex",
  "Keep(suffrDisplayClearPTV(\$pagename,PSS('$1')))");
$HTMLStylesFmt['clearptv'] = 'p.clearptv { background-color: #DDAADD; }';

function suffrDisplayClearPTV($pagename,$list)
{
  return "\n<p class='clearptv'><em>Clear on approval:</em> $list</p>";
}


/* (:topic name :) creates a new Wiki topic */
Markup('suffrtopic','fulltext',
  "/\\(:topic \\s+ (.+?) \\s* :\\)(.*?)\\(:topic \\s* :\\)/siex",
  "Keep(suffrDisplayTopic(\$pagename,PSS('$1'),PSS('$2')))");
$HTMLStylesFmt['topiclegend'] = 'p.topiclegend { background-color: #AABBDD; }';
$HTMLStylesFmt['topic'] = 'pre.topic { background-color: #BBCCEE; }';

function suffrDisplayTopic($pagename,$topicname,$body)
{
  $output = "\n<p class='topiclegend'><em>Create topic $topicname</em></p>";
  $output .= "\n<pre class='topic'>";
  $output .= $body;
  $output .= "</pre>\n";
  return $output;
}


/* (:removetopic name :) removes a topic */
Markup('suffrremovetopic','directives',
  "/\\(:removetopic \\s+ (.+?) \\s* :\\)/siex",
  "Keep(suffrDisplayRemoveTopic(\$pagename,PSS('$1')))");
$HTMLStylesFmt['removetopiclegend'] = 'p.removetopiclegend { background-color: #AABBDD; }';

function suffrDisplayRemoveTopic($pagename,$topicname)
{
  $output = "\n<p class='removetopiclegend'><em>Remove topic $topicname</em></p>";
  return $output;
}

SDV($suffrCmdFmt, "[[:alnum:]_\-\/]+"); //not complete yet.
SDV($suffrPageFmt, "[[:alnum:]]+\.[[:alnum:]]+"); //not complete yet.


/* (:cmdsig md5 exe :) stored exe signature on Stubs/CommandSignatures */
Markup('suffrcmdsig','directives',
  "/\\(:cmdsig \\s+ ($suffrMD5Fmt) \\s+ ($suffrCmdFmt) \\s* :\\)/siex",
  "Keep(suffrDisplayCmdSig(\$pagename,PSS('$1'),PSS('$2')))");
$HTMLStylesFmt['cmdsig'] = 'p.cmdsig { background-color: #887766; }';

function suffrDisplayCmdSig($pagename,$md5,$cmdpath) {
  $output = "\n<p class='cmdsig'><code>$cmdpath</code> $md5</p>";
  return $output;
}


/* (:pagesig:) displays page signature status on individual pages */
Markup('suffrpagesig','directives',
  "/\\(:pagesig:\\)/siex",
  "Keep(suffrDisplayPageSig(\$pagename))");
$HTMLStylesFmt['pagesig'] = 'p.pagesig {display: block; float: right; padding:0pt 3pt; font-size: 5pt;}';
$HTMLStylesFmt['pagesigy'] = 'p.psy { background-color: #8899AA; }';
$HTMLStylesFmt['pagesign'] = 'p.psn { background-color: #AA9988; }';

function suffrDisplayPageSig($pagename) {

  $opinion = suffrGetPageOpinion($pagename); 
  $output = "\n<p class='pagesig ps{$opinion['op']}'>Page status: {$opinion['opinion']}<!--<br>({$opinion['cur']})--></p>\n";
  return $output;
}


/* (:pagesig1 pagename md5 :) stored page signature in Stubs/PageSignatures */
Markup('suffrpagesig1','directives',
  "/\\(:pagesig1 \\s+ ($suffrPageFmt) \\s+ ($suffrMD5Fmt) \\s* :\\)/siex",
  "Keep(suffrDisplayPageSig1(\$pagename,PSS('$1'),PSS('$2')))");
$HTMLStylesFmt['pagesig1'] = 'p.pagesig1 { background-color: #8899AA; }';

function suffrDisplayPageSig1($pagename,$name,$md5) {
   return "\n<p class='pagesig1'>{$name}: {$md5}</p>\n";
}



/* Outstanding issues:
 * How to avoid multiple re-signings: edit the page w/ result to invalidate?
*/


/*
 * For redundancy and greater certainty, the author(s) _repeat_ that
 * they offer no warranty as to the safety or effectiveness of this
 * software.
 */
